/*
 *  platypus.c
 *
 *  Copyright 2010 Henry Kroll III www.thenerdshow.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */
 
#include "main.h"
#include "xkeys.h"
#include "runcommand.h"
/* Font selection dialog  */
VOID DIALOG_SelectFont(HWND hwnd) {
    CHOOSEFONTW cf;
    static LOGFONTW lf;
    HFONT getfont;
    HWND hEdit = GetDlgItem(hwnd, IDC_MAIN_EDIT);
    ZeroMemory(&cf, sizeof(cf));
    cf.lStructSize = sizeof(cf);
    cf.hwndOwner = hwnd;
    cf.lpLogFont = &lf;
    cf.Flags = CF_SCREENFONTS | CF_INITTOLOGFONTSTRUCT;
    if (ChooseFontW(&cf)) {
        getfont = CreateFontIndirectW(&lf);
        SendMessageW(hEdit, WM_SETFONT, (WPARAM) getfont, TRUE);
    }
}

/* Subclassed Edit WndProc handler  */
LRESULT CALLBACK EdWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam) {
    char *inCmdMsg;
#ifdef __WINE__
    static int markDelete,
    curpos=-1,oldsel,oldlParam;
    static char s[2]="@",Left[]={0x04,0},Right[]={0x06,0};
#endif
    int i,txtLen;
    switch (msg) {
    case EM_REPLACESEL:
        
        /*  Emulate Keypresses  */
        inCmdMsg=(char*)lParam;
/*      Is it a command?  */
        if(!cmdList(DO_COMMAND,inCmdMsg)){
#ifdef __WINE__
            //trim first space if replacing text
            if(oldsel&&inCmdMsg[0]==' ')inCmdMsg++;
            typeMsg(inCmdMsg);
            curpos+=(i=strlen(inCmdMsg));//cursor moves with msg
            debug("Replacing w=%i l=%s\n",wParam,inCmdMsg);
            if(oldsel){
                hkKey(" ");//space it out
                hkKey("\b");//delete it
                debug("Deleted\n");
            }else if(!i){
                debug("Delete that.\n");
                for(i=markDelete;i-->0;)
                    hkKey("\b");//really delete it
                curpos-=markDelete;
            }
            markDelete=i;
            oldsel=0;//selection is replaced!
#else
            printf("%s", inCmdMsg);
            fflush(stdout);
#endif
        } else return; /* suppress command output */        
        break;

    case EM_SETSEL:
        
        txtLen=SendMessage(hwnd,WM_GETTEXTLENGTH,0,0L);
        /* (re)initialize cursor */
        
#ifdef __WINE__
        if(curpos==-1)curpos=txtLen;
        debug("w %i l %i txtLen %i curpos %i oldsel %i\n", wParam,lParam,txtLen,curpos,oldsel);
        /* move cursor & select as lParam changes */
        if(lParam!=oldlParam){
            oldlParam=lParam;
            i=wParam-curpos;
            debug("moving %i spaces\n",i);
            if(i>0)s[0]=Right[0];//right arrow
            else s[0]=Left[0];//left arrow
            if((i=abs(i))){//if cursor moved
                if(oldsel)hkKey(Left);//break old selection
                while(i--)hkKey(s);//move it!
 
                s[0]=0x16;//highlight with shift arrow
                if((i=abs(lParam-wParam))){oldsel=i;
                    while(i--)hkKey(s);
                }else oldsel=0;
             }curpos=wParam;
        }
        /* "unselect that" */
        if(oldsel&&wParam==lParam){
            hkKey(Right);curpos=wParam;
            oldsel=0;
        }

        break;

    case WM_SETTEXT:
/*      Prevent app from going crazy g. */
        curpos=-1;oldsel=0;
        break;
#else
        break;
#endif
    case EM_SELECTALL:
        return CallWindowProc(g_edOldProc, hwnd, EM_SETSEL, 0, txtLen);
    }
/*  Pass control to default Proc  */
    return CallWindowProc(g_edOldProc, hwnd, msg, wParam, lParam);
}

#define LSW(IDS_APPNAME, appname) LoadStringW(GetModuleHandle(NULL), IDS_APPNAME, appname, sizeof(appname)/sizeof(WCHAR) );
/* Main WindProc handler  */
LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam) {
    HWND hEdit;
    hEdit = GetDlgItem(hwnd, IDC_MAIN_EDIT);
    char *t;
    long l;

    WCHAR About[] = { 'A','b','o','u','t',0 };
    WCHAR File[] = { '&','F','i','l','e', 0 };
    WCHAR Open[] = { '&','O','p','e','n', 0 };
    WCHAR Save[] = { '&','S','a','v','e', 0 };
    WCHAR SaveAs[] = { 'S','a','v','e','&','A','s', 0 };
    WCHAR Exit[] = { 'E','&','x','i','t', 0 };
    WCHAR New[] = { '&','N','e','w', 0 };
    WCHAR Edit[] = { '&','E','d','i','t', 0 };
    WCHAR Font[] = { '&','F','o','n','t', 0 };
    WCHAR Undo[] = { '&','U','n','d','o', 0 };
    WCHAR Cut[] = { 'C','u','&','t',0 };
    WCHAR Copy[] = { '&','C','o','p','y', 0 };
    WCHAR Paste[] = { '&','P','a','s','t','e', 0 };
    WCHAR SelectAll[] = { '&','S','e','l','e','c','t',' ','A','l','l', 0 };
    
    WCHAR Actions[] = { '&','A','c','t','i','o','n','s', 0 };
    WCHAR Erase[] = { 'E','&','r','a','s','e',' ','A','l','l', 0 };
    WCHAR Pasteit[] = { '&','P','a','s','t','e',' ','a','n','y','w','h','e','r','e','.','.','.', 0 };
    WCHAR Help[] = { '&','H','e','l','p', 0 };

    switch (msg) {

    case WM_CREATE:
        {
            HMENU hMenu, hSubMenu;

            hMenu = CreateMenu();       /* Create a menu bar at the top */

            hSubMenu = CreatePopupMenu();
            AppendMenuW(hMenu, MF_STRING | MF_POPUP, (UINT_PTR) hSubMenu, File);
            AppendMenuW(hSubMenu, MF_STRING, IDM_FILE_OPEN, Open);
            AppendMenuW(hSubMenu, MF_STRING, IDM_FILE_SAVE, Save);
            AppendMenuW(hSubMenu, MF_STRING, IDM_FILE_SAVEAS, SaveAs);
            AppendMenuW(hSubMenu, MF_STRING, IDM_FILE_EXIT, Exit);
            AppendMenuW(hSubMenu, MF_STRING, IDM_ACTIONS_SETTEXT, New);

            hSubMenu = CreatePopupMenu();
            AppendMenuW(hMenu, MF_STRING | MF_POPUP, (UINT_PTR) hSubMenu, Edit);
            AppendMenuW(hSubMenu, MF_STRING, IDM_EDIT_SETFONT, Font);
            AppendMenuW(hSubMenu, MF_STRING, IDM_EDIT_UNDO, Undo);
            AppendMenuW(hSubMenu, MF_STRING, IDM_EDIT_CUT, Cut);
            AppendMenuW(hSubMenu, MF_STRING, IDM_EDIT_COPY, Copy);
            AppendMenuW(hSubMenu, MF_STRING, IDM_EDIT_PASTE, Paste);
            AppendMenuW(hSubMenu, MF_STRING, IDM_EDIT_SELECTALL, SelectAll);

            hSubMenu = CreatePopupMenu();
            AppendMenuW(hMenu, MF_STRING | MF_POPUP, (UINT_PTR) hSubMenu, Actions);
#ifdef __WINE__
            AppendMenuW(hSubMenu, MF_STRING, IDM_ACTIONS_PASTEIT, Pasteit);
#endif
            AppendMenuW(hSubMenu, MF_STRING, IDM_ACTIONS_SETTEXT, Erase);

            hSubMenu = CreatePopupMenu();
            AppendMenuW(hMenu, MF_STRING | MF_POPUP, (UINT_PTR) hSubMenu, Help);
            AppendMenuW(hSubMenu, MF_STRING, IDM_HELP_ABOUT, About);

            SetMenu(hwnd, hMenu);
        }
        break;

    case WM_COMMAND:
        {
/*          LOWORD handles accelerators in the same loop */
            wParam=LOWORD(wParam);
            switch (wParam) {
            case IDM_FILE_EXIT:
                debug("closing...\n");
                if(fileFun(hwnd,IDM_FILE_QSAVE)!=IDCANCEL)
                    PostQuitMessage(0);
                else {
                    SetForegroundWindow(hwnd);
                    return;
                }
                break;
            case IDM_FILE_OPEN:
                fileFun(hwnd,wParam);
                break;
            case IDM_FILE_SAVE:
                fileFun(hwnd,wParam);
                break;
            case IDM_FILE_SAVEAS:
                fileFun(hwnd,wParam);
                break;
            case IDM_EDIT_SETFONT:
                DIALOG_SelectFont(hwnd);
                break;
            case IDM_EDIT_CUT:
                SendMessage(hEdit, WM_CUT, 0, 0);
                break;
            case IDM_EDIT_COPY:
                SendMessage(hEdit, WM_COPY, 0, 0);
                break;
            case IDM_EDIT_PASTE:
                SendMessage(hEdit, WM_PASTE, 0, 0);
                break;
            case IDM_EDIT_SELECTALL:
                SendMessage(hEdit, EM_SELECTALL, 0, 0);
                break;
            case IDM_EDIT_UNDO:
                if(SendMessage(hEdit, EM_CANUNDO, 0, 0)!=0)
                    SendMessage(hEdit, EM_UNDO, 0, 0);
                break;
#ifdef __WINE__
            case IDM_ACTIONS_PASTEIT:
                l=SendMessage(hEdit, WM_GETTEXTLENGTH, 0, 0);
                if(!l){
                    MessageBox(hwnd,"You would have to type some text first...",
                "Empty buffer!",MB_OK);
                    return;
                }
                if(MessageBox(hwnd,"Select app to send text to. You will have 3 seconds.",
                "Send text...",MB_OKCANCEL)==IDCANCEL)break;
                sleep(3);
                t=malloc(l+2);
                if(!t) ERM("Out of memory!");
                GetDlgItemText(hwnd, IDC_MAIN_EDIT, t,l+1);
                typeMsg(t);
                free(t);
                break;
#endif
            case IDM_ACTIONS_SETTEXT:
                if(SendMessage(hEdit, EM_GETMODIFY, 0, 0)&&
                fileFun(hwnd,IDM_FILE_QSAVE)==IDCANCEL){
                    SetForegroundWindow(hwnd); break;
                }
                else fileFun(hwnd,IDM_FILE_NOSAVE);
                SetDlgItemText(hwnd, IDC_MAIN_EDIT, "");
                SetForegroundWindow(hwnd); break;
            case IDM_HELP_ABOUT:
                {
                    WCHAR appname[99], about[256];
                    LSW(IDS_APPNAME, appname);
                    LSW(IDS_ABOUT, about);
                    ShellAboutW( hwnd, appname, about,
                    LoadImageA(GetModuleHandle(NULL), "PLATYPUS",
                    IMAGE_ICON, 48, 48, LR_SHARED ));
                    return 0;
                }break;
            }return 0;
        }break;
    case WM_SETFOCUS:
        {
            SetFocus(hEdit);
        }
        break;
    case WM_SIZE:
        {
            RECT rcClient;
            GetClientRect(hwnd, &rcClient);
            SetWindowPos(hEdit, NULL, 0, 0, rcClient.right, rcClient.bottom,
                         SWP_NOZORDER);
        }
        break;
    case WM_DESTROY:
        {
            PostQuitMessage(0);
        }
        break;

    case WM_CLOSE:
        {
            if(fileFun(hwnd,IDM_FILE_QSAVE)!=IDCANCEL)
                PostQuitMessage(0);
            else {
                SetForegroundWindow(hwnd);
                return;
            }
        }
        break;

default:
        return DefWindowProc(hwnd, msg, wParam, lParam);
    }
    return 0;
}

int WINAPI
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow) {
    WNDCLASSEX wc, ed;
    int i;
    HWND hwnd;
    MSG Msg;
    char appname[25];
    HACCEL haccel;
    
    LoadString(hInstance, IDS_APPNAME, appname, sizeof(appname));
#ifdef __WINE__
    readKeysFile("platypus_keys");
    openDisplay(lpCmdLine);
#endif
    cmdList(READ_COMMAND_FILE,"platypus_commands");
    
/*  Subclass the desktop window: get default settings */
    GetClassInfoEx(0, "#32769", &wc);
    wc.cbSize           = sizeof (WNDCLASSEX);
    wc.lpfnWndProc      = WndProc;
    wc.hInstance        = hInstance;
    wc.lpszClassName    = appname;
    wc.hIcon = LoadIcon( hInstance, "PLATYPUS" );

/*  Subclass the edit control: get default settings */
    GetClassInfoEx(0, "edit", &ed);
    ed.cbSize           = wc.cbSize;
    g_edOldProc = ed.lpfnWndProc; /* Save default WndProc */
    ed.lpfnWndProc = EdWndProc;   /* Hook our own WndProc */
    ed.hInstance = hInstance;
    ed.lpszClassName = "hkEDIT";  /* Our custom class name */

/*  Register the classes  */
    if (!RegisterClassEx(&wc))
        ERM("Window Registration Failed!");
    if (!RegisterClassEx(&ed))
        ERM("Edit Box Registration Failed!");
/*  Create a window  */
    if (!(hwnd = CreateWindowEx(WS_EX_CLIENTEDGE|WS_EX_TOPMOST,
                                appname,
                                appname,
                                WS_OVERLAPPEDWINDOW,
                                CW_USEDEFAULT, CW_USEDEFAULT,640,460,
                                NULL, NULL, hInstance, NULL)))
        ERM("Window Creation Failed!");
    HWND hEdit;
/*  Drop in an edit control  */
    if (!(hEdit = CreateWindowEx(WS_EX_CLIENTEDGE, "hkEDIT", "",
                                WS_CHILD | WS_VISIBLE | WS_VSCROLL |
                                ES_MULTILINE | ES_AUTOVSCROLL | ES_WANTRETURN, 0, 0,
                                100, 100, hwnd, (HMENU) IDC_MAIN_EDIT,
                                hInstance, NULL)))
        ERM("Could not create edit box.");

    ShowWindow(hwnd, SW_SHOWNORMAL);
    UpdateWindow(hwnd);
   
    if(!(haccel=LoadAccelerators(hInstance, "EditAccel")))
        ERM("Could not load string accelerator table."); 


/*  Main Message Loop  */
    while (GetMessage(&Msg, (HWND)NULL, 0, 0) > 0) {
        if (!TranslateAccelerator(hwnd,haccel,&Msg)){
            TranslateMessage(&Msg);
            DispatchMessage(&Msg);
        }
    }
#ifdef __WINE__
    if (xD) {
/*      Close X display */
        XTestDiscard(xD);
        XFlush(xD);
        XCloseDisplay(xD);
    }
#endif
/*  Free up command memory  */
    cmdList(CLEANUP,NULL);
    
    return Msg.wParam;
}
