/*
 *  runcommand.c
 *
 *  Copyright 2010 Henry Kroll III www.thenerdshow.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */
#include "main.h"
#include "xkeys.h"
#include "runcommand.h"

/*  Count lines in a file  */
int countLines(FILE *f) {
    int i=0;
    while(fscanf(f," %*2048[^\n]")!=EOF)i++;
    rewind(f); return i;
}

/*  Command List Object  */
BOOL cmdList(METHOD m,char *inCmdMsg) {
    FILE *f;int i=0;
    static long fsize;
    static char **inCmd,**outCmd;
    char erm[300],*oc;
    char s[255];
    switch(m){
        case READ_COMMAND_FILE:

#ifdef __WINE__
            strcpy(s,winePath(INSTALL_PATH));
#else
            strcpy(s,INSTALL_PATH);
#endif
            strncat(s,inCmdMsg,255);
            sprintf(erm, "File %s not found!", s);
            if(!(f=fopen(s,"r")))ERM(erm);
            fsize=countLines(f);
            inCmd=malloc(fsize*sizeof inCmd);
            if(!inCmd) ERM("Out of memory!");
            outCmd=malloc(fsize*sizeof inCmd);
            if(!outCmd) ERM("Out of memory!");
            for(i=fsize;i--;){
                if(!(inCmd[i]=malloc(99)))ERM("Out of memory!");
                if(!(outCmd[i]=malloc(99)))ERM("Out of memory!");
            }                
            for(i=fsize;i--;)
                fscanf(f," \"%99[^\"]\" \"%99[^\"]\" ",inCmd[i],outCmd[i]);
            if(DEBUG)
                for(i=fsize;i--;)debug("(%s)(%s)\n",inCmd[i],outCmd[i]);
        break;
        case DO_COMMAND:
            
            if(StrStrI(inCmdMsg,"reset dictation")){
                SetDlgItemText(GetForegroundWindow(), IDC_MAIN_EDIT, "");
                return TRUE;
                
            }else if(StrStrI(inCmdMsg,"undo all that")){
                HWND hEdit;
                hEdit = GetDlgItem(GetForegroundWindow(), IDC_MAIN_EDIT);

                if(SendMessage(hEdit, EM_CANUNDO, 0, 0)!=0)
                    SendMessage(hEdit, EM_UNDO, 0, 0);
#ifdef __WINE__              
                hkKeySt("Control_L", KEYDOWN);
                hkKeySt("z",KEYDOWN);
                hkKeySt("z",KEYUP);
                hkKeySt("Control_L", KEYUP);
#endif  
                return TRUE;
                
            }else for(i=fsize;i--;){
                if(StrStrI(inCmdMsg,inCmd[i])){
                /*  Run command if it ends in &  */
                    if(outCmd[i][strlen(outCmd[i])-1]=='&'){
                        debug("COMMAND: %s\n",inCmdMsg);
                        system(outCmd[i]);
                    }else{
                 /* Otherwise interpret it as keystrokes */
                    strcpy(erm,outCmd[i]);
                    for(oc=strtok(erm," ");oc
                        ;oc=strtok(NULL," ")){
#ifdef __WINE__
                        /* Hold down any modifier keys */
                        if(oc[1]){
                            hkKeySt(oc,KEYDOWN);
                            debug("KEYDOWN: %s\n",oc);
                        }else{
                         /* Fake a keypress */
                            debug("%s\n",oc);
                            hkKey(oc);
                        }
#else
/*  Print text to terminal. Treat all modifiers as spaces.  */
                        if(oc[1])strcpy(oc," ");
                        printf("%s",oc);
#endif                      
                    }
#ifdef __WINE__
/*                  Release modifier keys  */
                    strcpy(erm,outCmd[i]);
                    for(oc=strtok(erm," ");oc
                        ;oc=strtok(NULL," ")){
                            if(oc[1]){
                                hkKeySt(oc,KEYUP);
                                debug("KEYUP: %s\n",oc);
                            }
                        }
#else
                    printf("\n");
#endif
                    }
                    return TRUE;
                }
            }
            return FALSE;
            
        case CLEANUP:

            for(i=fsize;i--;)free(inCmd[i]),free(outCmd[i]);
            free(inCmd); free(outCmd);

        break;
    }
    return FALSE;
}
