/*
 *  xkeys.c
 *
 *  Copyright 2010 Henry Kroll III www.thenerdshow.com
 *
 * These keyboard macro sections were vaguely inspired by
 * Xmacro (xmacro.sourceforge.net) which is:
 * Copyright (C) 2000 Gabor Keresztfalvi <keresztg@mail.com>
 *
 * See README for additional details.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * WinePath is necessary when linking against Linux libc instead of
 * windows' or wine's msvcrt. But linux's libc lets us launch Linux apps
 * from within our program instead of just Windows apps.
 */

#include "main.h"
#include "xkeys.h"
#ifdef __WINE__

char * winePath(char *inPath) {
    static char s[MAX_PATH]={0};
    FILE *f;
    int nmeb;
/*  winepath 'inPath'  */
    strcpy(s,"winepath '");strcat(s,inPath);strcat(s,"'");
    if(!(f=popen(s,"r")))ERM("winepath error");
    nmeb=fread(s,1,MAX_PATH,f);pclose(f);
    if(nmeb>1)s[nmeb-1]='\0';
    return s;
}

/*     Read keys file  */
VOID readKeysFile(char *inFile) {
    FILE *f;
    char s[255];
    unsigned char i = 0;
    char erm[300];
    
    strcpy(s,winePath(INSTALL_PATH));
    strcat(s,"/");
    strncat(s,inFile,255);
    sprintf(erm,"File %s not found!",s);
    if (!(f = fopen(s, "r")))
        ERM(erm);
    while (!feof(f)){
        fscanf(f, "%i", &i);
        fscanf(f, " %99s", g_keys[i&=0xff]);
        fscanf(f, "%99[^\n]", s);
        /* fprintf(stderr, "%x %s\n",i, g_keys[i]); */
    }
    fclose(f);
}

/* Open X Display  */
VOID openDisplay(const char *DisplayName) {
    int Event, Error;
    int Major, Minor;
    KeySym ks;
    KeyCode kc;
    char erm[99];
    if (DisplayName[0])
        xD = XOpenDisplay(DisplayName);
    else
        xD = XOpenDisplay(":0.0");
    if (!xD){
        sprintf(erm, "Could not open display %s", DisplayName);
        ERM(erm);
    }
/*  Fail on no Xtest-extension  */
    if (!XTestQueryExtension(xD, &Event, &Error, &Major, &Minor)){
        XCloseDisplay(xD);
        ERM("X server does not support the Xtest extension.");
    }
/*  Grab Control  */
    XTestGrabControl(xD, True);
    XSync(xD, True);
    DefaultScreen(xD);
    XTestDiscard(xD);
}

/*  Press or release a key. */
VOID hkKeySt(char *s, BOOL ud) {
    KeyCode kc;
    KeySym ks = XStringToKeysym(s);
    if ((kc = XKeysymToKeycode(xD, ks)) != 0){
        XTestFakeKeyEvent(xD, kc, ud, Delay);
        /* fprintf(stderr, "  %s '%s' %x %x\n",(ud)?"press":"release",s,ks,kc); */
        XFlush(xD);
    }
}

/*  Generate a fake keypress  */
VOID hkKey(unsigned char *s) {
    BOOL shift = False;
    char shifted[]="  ~!@#$%^&*()<>_+|}{\":?>ABCDEFGHIJKLMNOPQRSTUVWXYZ";
    char hex[3]="FF";
    /* implement shifted arrow keys so Platypus can select text */
    if(shifted[0]==' '){shifted[0]=0x14;shifted[1]=0x16;}
    
    /* debug show keycodes, keysyms ref. /usr/include/X11/keysymdef.h */
    /* fprintf(stderr, "%05.2x %s\n",s[0], XKeysymToString((KeySym)*s)); */
    /* test with ¡¢£¤¥¦§¨©«¬ ®¯²³´µ¶¹»¼½¾¿ÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖ×ØÙÚÛÜÝÞß */

    /* if keycode is unicode or not defined */
    if (!g_keys[s[0]][0]||strstr(g_keys[s[0]],"Ux")){
        /* type Ctrl-Shift-u for ISO 14755-conformant input (GTK+) */
        hkKeySt("Control_L", KEYDOWN);
        hkKeySt("Shift_L", KEYDOWN);
        hkKeySt("u", KEYDOWN);
        hkKeySt("u", KEYUP);
        hkKeySt("Shift_L", KEYUP);
        hkKeySt("Control_L", KEYUP);    
        /* type out hex keycode from platypus_keys file */
        if (g_keys[s[0]][0])
            typeMsg(g_keys[s[0]]+2);        
        else { /* guess keycode for other keys */
            /* usually it's the same as the key number in hex */
            sprintf(hex,"%0.1x%0.1x",s[0]/16,s[0]&15);
            typeMsg(hex);
        } return;
        
    /* support AltGr keyboard layouts see README */
    /* copy platypus_keys.altgr to platypus_keys */
    } else if (strstr(g_keys[s[0]],"A_")){
        hkKeySt("ISO_Level3_Shift", KEYDOWN);
        /* hack wine keyboard layout */
        if (*(g_keys[s[0]]+2)=='<') {
            hkKeySt("Shift_R", KEYDOWN);
            hkKeySt("Ccedilla",KEYDOWN);
            hkKeySt("Ccedilla",KEYUP);
            hkKeySt("Shift_R", KEYUP);
        } else if (*(g_keys[s[0]]+2)=='`') {
            hkKeySt("dead_grave",KEYDOWN);
            hkKeySt("dead_grave",KEYUP);
        } else if (*(g_keys[s[0]]+2)=='\'') {
            hkKeySt("dead_acute",KEYDOWN);
            hkKeySt("dead_acute",KEYUP);
        } else
            hkKey(g_keys[s[0]]+2);
        /* hold AltGr for ¨ */
        if (*(g_keys[s[0]]+2) == *(g_keys[s[0]]+3)){
            hkKey(g_keys[s[0]]+3);
            hkKeySt("ISO_Level3_Shift", KEYUP);
        /* otherwise lift up the alt key and type the rest */
        } else {
            hkKeySt("ISO_Level3_Shift", KEYUP);
            typeMsg(g_keys[s[0]]+3);
        }
    
    } else {
        /* look up keycode in platypus_keys file */
        /* press shift if it's a capital letter */
        if (strchr(shifted, s[0])){
            shift = True;
            hkKeySt("Shift_R", KEYDOWN);
        } /* press and release the keys */
        hkKeySt(g_keys[s[0]],KEYDOWN);
        hkKeySt(g_keys[s[0]],KEYUP);
        if (shift) hkKeySt("Shift_R", KEYUP);
    }
}

VOID typeMsg(char *inCmdMsg) {
    char s[5] = "@";
    int i = 0;
    while (inCmdMsg[i]){
        sprintf(s,"%c",inCmdMsg[i]);
        hkKey(s);i++;
    }
}

#endif